'''
Created on 2010-07-19

@author: Shobhit Jain

@contact: shobhit@cs.toronto.edu
'''

from ontology import GOGraph

def load_semantic_similarity(ontology_file, gene_file, ontology, code):
    '''
    Calls functions for loading and processing data.
    '''
    objs = {}
    ontology = ontology.split(",")
    g = GOGraph()
    print "loading data............."
    g._obo_parser(ontology_file)
    g._go_annotations(gene_file, code)
    print "processing data..........(takes time)"
    run = {'C':g._cellular_component, 'P':g._biological_process, 'F':g._molecular_function}
    for i in ontology:
        i = i.split(":")
        objs[i[0]] = run[i[0]]()
        objs[i[0]]._species()
        objs[i[0]]._clustering(float(i[1]))
    return objs



def return_detail(result, geneA, geneB, detail):
    '''
    Formats the output for printing on screen or on file.
    '''
    domain_def = {'C':'Cellular Component', 'P':'Biological Process', 'F':'Molecular Function'}
    r = "\nSemantic similarity between " + geneA + " and " + geneB + " is:\n\n"
    for domain in result:
        r += " " + domain_def[domain] + ": " + str(result[domain][0])
        if detail:
            for data in result[domain][1]:
                r += "\n  GO id assigned to " + geneA + " is: " + data[0] + \
                     "\n  GO id assigned to " + geneB + " is: " + data[1] + \
                     "\n  LCA of assigned GO ids is: " + "|".join(result[domain][1][data]['lca']) + "\n\n"                  
    return r + "\n\n\n"
        
        

def calculate_semantic_similarity(objs, geneA, geneB, detail):
    '''
    Calls the function for calculating semantic similarity between
    genesA and genesB.
    '''
    result = {}
    for domain in objs:
        result[domain] = objs[domain]._semantic_similarity(geneA, geneB)
    return return_detail(result, geneA, geneB, detail)
        
    #cc._semantic_similarity('S000004065', 'S000001451')
    



'''
Created on 2010-07-26

@author: Shobhit Jain

@contact: shobhit@cs.toronto.edu
'''

class Parser(object):
    '''
    Parser class implement functions for parsing input data files.
    '''
    
    def __init__(self):
        '''
        Constructor
        '''
        super(Parser, self).__init__()
        
        
     
    def _obo_parser(self, obo_file):    
        '''
        Parser for Gene Ontology obo files. go_annotations variable is 
        updated.
        '''
        file = open(obo_file, 'r')
        flag = 0
        for line in file:
            line = line.strip()
            if line.startswith('[Term]'): 
                flag = 1
                node = ''
                name = ''
                domain = ''
                parent = set()
            elif flag == 1 and line == '': 
                flag = 0
                self._add_node(node)
                for term in parent:
                    self._add_node(term)
                    self._add_edge(term, node)
                self.go_annotations[node] = {'name':name, 
                                             'domain':domain, 
                                             'gene':set(), 
                                             'ancestors':set(), 
                                             'cluster':{}
                                             }
            elif flag == 1 and line.startswith("id"):
                node = line.split('id:')[1].strip()   
            elif flag == 1 and line.startswith("namespace"):
                domain = line.split('namespace:')[1].strip()
            elif flag == 1 and line.startswith("name"):
                name = line.split('name:')[1].strip()
            elif flag == 1 and line.startswith("is_a"):
                parent.add(line.split(' ')[1])
            elif flag == 1 and line.startswith("relationship"):
                parent.add(line.split(' ')[2])
                
                
                
    def _go_annotations(self, gene_file, cd):
        '''
        Parser for gene annotation file (SGD/human). go_annotations variable is updated.
        '''
        file = open(gene_file, 'r')
        for line in file:
            line = line.strip()
            if line.startswith('!'):
                continue
            elif line.startswith('SGD') or line.startswith('UniProtKB'): #change this line to read data from annotation file
                line = line.split('\t')
                term = line[4].strip()
                gene = line[1].strip()
                code = line[6].strip()
                if term in self.go_annotations and code != cd:
                    self.go_annotations[term]['gene'].add(gene)
                    
                
                
                

                        
                        



'''
Created on 2010-07-26

@author: Shobhit Jain

@contact: shobhit@cs.toronto.edu
'''

class SemanticSimilarity(object):
    '''
    SemanticSimilarity class implement functions for calculating 
    topology based semantic similarity between genes.  
    '''


    def __init__(self):
        '''
        Constructor
        '''
        super(SemanticSimilarity, self).__init__()
        
        
    def _lowest_common_ancestor(self, termA, termB, cluster):
        '''
        Given GO terms A and B and the graph to which they belong
        the function finds the lowest common ancestor i.e. the 
        ancestor of terms A and B with highest entropy. 
        '''
        lca = {}
        for term in self.go_annotations[termA]['cluster'][cluster]['ancestors'] \
                            .intersection(self.go_annotations[termB]['cluster'][cluster]['ancestors']):
            entropy = self.go_annotations[term]['cluster'][cluster]['entropy']
            if entropy not in lca:
                lca[entropy] = set()
            lca[entropy].add(term) 
        return max(lca), lca[max(lca)]
        
        
        
    def _semantic_similarity(self, geneA, geneB):
        '''
        Given genes A and B the function calculates the semantic similarity
        between them. It uses gene_annotations and go_annotaions variables.
        '''
        sem_sim = {}
        for termA in self.gene_annotations[geneA]:
            for termB in self.gene_annotations[geneB]:
                for clusA in self.go_annotations[termA]['cluster']:
                    for clusB in self.go_annotations[termB]['cluster']:
                        if clusA != 'meta' and clusB != 'meta' and clusA == clusB:
                            value, lca = self._lowest_common_ancestor(termA, termB, clusA)
                            if value not in sem_sim:
                                sem_sim[value] = {}
                            sem_sim[value][(termA, termB)] = {'lca':lca, 'clusA':clusA, 'clusB':clusB}
                        elif clusA != 'meta' and clusB != 'meta' and clusA != clusB:
                            value, lca = self._lowest_common_ancestor(clusA, clusB, 'meta')
                            if value not in sem_sim:
                                sem_sim[value] = {}
                            sem_sim[value][(termA, termB)] = {'lca':lca, 'clusA':clusA, 'clusB':clusB}
        return max(sem_sim), sem_sim[max(sem_sim)]
                            
                            
                
            


'''
Created on 2010-07-27

@author: Shobhit Jain

@contact: shobhit@cs.toronto.edu
'''

import sys
import os
import getopt
from main import load_semantic_similarity, calculate_semantic_similarity


def _usage():
    '''
    Details on how to use TCSS.
    '''
    print "\n\n tcss.py [-options] geneA geneB\
    \n or \n tcss.py [-options] -i input_file"
    print "\n -options \
    \n    -i [file name] or --input [=file name]       Input file (two genes separted by comma per line)\
    \n    -o [file name] or --output [=file name]      Output file\
    \n    -c [domain:cutoff] or                        Domain [C/P/F], cutoff [int/float] in any combination\
    \n         --topology-cutoff [=domain:cutoff]      (default: C:2.4,P:3.5,F:3.3)\
    \n    --detail                                     Detailed output (default: False)\
    \n    --gene [=file name]                          Gene annotation file (default: SGD file provided)\
    \n    --go [=file name]                            Gene Ontology (GO) obo file (default: GO file provided)\
    \n    --drop [=evidence code]                      GO evidence code not to be used \
    \n    -h or --help                                 Usage\n\n\n"



def _command_line_arguments(argv):
    '''
    Process the command line arguments. Returns options variable.
    '''
    options = {'args':None, 'input':None, 'output':None, 'detail':None, 'ontology':"C:2.4,P:3.5,F:3.3", \
               'gene':'gene_association.sgd', 'go':'gene_ontology.obo.txt', 'drop':''}
    try:
        opts, args = getopt.getopt(argv, "hi:o:c:", ["help", "input=", "output=", "topology-cutoff=", \
                                                     "detail", "ontology=", "gene=", "go=", "drop="])
    except getopt.GetoptError:
        _usage()
        sys.exit(2)
    if len(opts) == len(args) == 0:
        _usage()
        sys.exit()
    if 1 < len(args) < 3:
        options['args'] = args
    if len(args) > 2:
        _usage()
        sys.exit()
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            _usage()
            sys.exit()
        elif opt in ("-i", "--input"):
            if os.path.isfile(arg):
                if len(args) > 0: 
                    print "Excluding arguments:", args
                    options['args'] = None
                options['input'] = arg
            else:
                print 'Input file not found'
                sys.exit()
        elif opt in ("-o", "--output"):
            options['output'] = arg
        elif opt in ("-c", "--ontology"):
            options['ontology'] = arg
        elif opt in ("--detail"):
            options['detail'] = True
	elif opt in ("--drop"):
	    options['drop'] = arg
        elif opt in ("--gene"):
            if os.path.isfile(arg):
                options['gene'] = arg
            else:
                print 'Gene annotation file not found'
                sys.exit()
        elif opt in ("--go"):
            if os.path.isfile(arg):
                options['go'] = arg
            else:
                print 'GO file not found'
                sys.exit()
        else:
            _usage()
            sys.exit()     
    if options['args'] == None and options['input'] == None:
        _usage()
        sys.exit()
    return options



def input_file_run(objs, infile, outfile, detail):
    '''
    Run TCSS when input file is provided.
    '''
    file = open(infile, 'r')
    result = ""
    for line in file:
        line = line.strip().split(",")
        if len(line) > 2:
            print 'File format not proper: refer README'
            sys.exit()
        result += calculate_semantic_similarity(objs, line[0], line[1], detail)
    file.close()
    output_results(result, outfile)     
        
        
        
def output_results(result, outfile):
    '''
    Output the results on screen or on file. 
    '''
    if outfile:
        try:
            file = open(outfile, 'w')
            file.write(result)
            file.close()
        except:
            print "Cannot open output file: check path or permissions \n Printing results on screen"
            print result
    else:
        print result
        

    
if __name__ == '__main__':
    
    options = _command_line_arguments(sys.argv[1:])
    objs = load_semantic_similarity(options['go'], options['gene'], options['ontology'], options['drop'])
    
    if options['input']:
        input_file_run(objs, options['input'], options['output'], options['detail'])
    else:
        result = calculate_semantic_similarity(objs, options['args'][0], options['args'][1], options['detail'])
        output_results(result, options['output'])
    



'''
Created on 2010-07-20

@author: Shobhit Jain

@contact: shobhit@cs.toronto.edu
'''

from graph import Graph
from clustering import Clustering
from parser import Parser
from semantic import SemanticSimilarity

class GOGraph(Graph, Clustering, Parser, SemanticSimilarity):
    '''
    Class GOGraph implement graph functions specific for
    gene ontology and inherits Graph, Clustering, Parser,
    SemanticSimilarity classes.
    '''


    def __init__(self):
        '''
        Initialises go_annotations, goid, and gene_annotations 
        variables.
        
        go_annotations variable contains detailed information of
        GO terms and also stores the results of data processing to
        be used for calculating semantic similarity.
        {GO_term:{'name':"",'domain':"", 'gene':set(), 'ancestors':set(),
         'clusters':{'cluster_id':"", 'entropy':float(), 'ancestors':set()
         }}}
        
        goid takes one of the following values: 
        GO:0005575 (cellular_component)
        GO:0008150 (biological_process)
        GO:0003674 (molrcular_function)
        
        gene_annotations variable contains GO terms assigned to genes. 
        {gene:set(GO_term)}
        '''
        super(GOGraph, self).__init__()
        self.go_annotations = {}
        self.goid = ''
        self.gene_annotations = {}
          
        
        
    def _create_graph_from_node_set(self, node_set):
        '''
        Given a set of nodes create a graph from it. GOGraph object is
        returned.
        '''
        new_graph = GOGraph()
        for node1 in node_set:
            new_graph._add_node(node1)
            for node2 in node_set:
                if node1 != node2 and self._depth_first_search(node = node1, target = node2):
                    new_graph._add_edge(node1, node2)
        new_graph._remove_unwanted_edges()
        new_graph.go_annotations = self.go_annotations
        new_graph.goid = self.goid
        return new_graph
                                
                    
                    
    def _create_graph_from_node(self, node):
        '''
        Given a node create use depth first search to create a new graph.
        GOGraph object is returned.
        '''
        new_graph = GOGraph()
        #new_graph.go_annotations = self.go_annotations
        new_graph.goid = node
        new_terms =  self._depth_first_search(node)
        new_graph._update_graph(new_terms)
        for term in new_terms:
            new_graph.go_annotations[term] = self.go_annotations[term]
        return new_graph
                                
               
                                
    def _species(self):
        '''
        Prune the current graph to have nodes with genes annotated to them for
        a particular species. For this genes are propagated first to parent
        terms.
        '''
        for node in self.flip:
            parent_terms = self._depth_first_search(node = node, rev = True).keys()
            self.go_annotations[node]['ancestors'] = set(parent_terms)
            for parent in parent_terms:
                self.go_annotations[parent]['gene'] = self.go_annotations[parent]['gene'] \
                                                    .union(self.go_annotations[node]['gene'])
        for term in self.graph.keys():
            if len(self.go_annotations[term]['gene']) == 0:
                self.go_annotations.pop(term)
                for parent in list(self.flip[term])[:]:
                    self._remove_edge(parent, term)
                self._remove_node(term)    
                
                
                
    def _gene_annotations(self):
        '''
        Assign genes to the most specific node among all the nodes it has been 
        originally annotated to in a graph. 
        '''
        for term in self.go_annotations:
            for gene in self.go_annotations[term]['gene']:
                if gene not in self.gene_annotations:
                    self.gene_annotations[gene] = set()
                self.gene_annotations[gene].add(term) 
            
        for gene in self.gene_annotations:
            go_list = list(self.gene_annotations[gene])[:]
            for term1 in go_list:
                for term2 in go_list:
                    if term1 != term2 and \
                        term1 in self.go_annotations[term2]['ancestors'] \
                        and term1 in self.gene_annotations[gene]:
                        self.gene_annotations[gene].remove(term1)
                        
            
                
                    
    def _clustering(self, cutoff):
        '''
        Function takes the topology cutoff as the input and call other functions
        to create higher/lower level graphs and entropy calculation. go_annotation
        and gene_annotation variables are updated will all the information.
        '''
        meta_terms = self._meta_graph_nodes(cutoff)
        meta_graph = self._create_graph_from_node_set(meta_terms)
        self._entropy_graph(meta_graph, 'meta')
        self._ancestor_terms(meta_terms, 'meta')
        sub_terms = self._sub_graph_nodes(meta_graph, meta_terms)
        sub_graph =  {}
        for term in sub_terms:
            sub_graph[term] = self._create_graph_from_node_set(sub_terms[term])
            self._entropy_graph(sub_graph[term], term)
            self._ancestor_terms(sub_terms[term], term)
        self._gene_annotations()    
        
           
                    
    def _cellular_component(self):
        '''
        Creates a cellular component gene ontology graph.
        '''
        return self._create_graph_from_node('GO:0005575')
    
    
    
    def _biological_process(self):
        '''
        Creates a biological process gene ontology graph.
        '''
        return self._create_graph_from_node('GO:0008150')
    
    
    
    def _molecular_function(self):
        '''
        Creates a molecular function gene ontology graph.
        '''
        return self._create_graph_from_node('GO:0003674')
    
    
    
                    
                    
                
            
                
                                
                